#include "combine.h"
#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"
#include <unistd.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <sys/time.h>

/* 
 * Combine operation implementation
 *
 * Author: Wi Bing Tan
 *
 * Description: This program runs all-gather, all-reduce and 
 * reduce-scatter MPI combine operations using customised 
 * algorithms. The output is the time taken to run the 
 * algorithm.
 */

int main(argc, argv)
     int argc;
     char **argv;  
{
  int rank;
  int size;
  int error;
  struct args *args;
  double *sendbuf, *recvbuf;
  int done;
  int count;
  double test;
  int i, j;
  int *recvcounts;
  int numrecvelem;
  struct timeval *tv1, *tv2; 
  struct timezone *tz;
  long timediff;
  long avgtime=-1;
  long maxtime=-1;
  long mintime=LONG_MAX;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank );
  MPI_Comm_size(MPI_COMM_WORLD, &size );

  // allocate argument structure and time values

  if ((args = (struct args*)malloc(sizeof(struct args))) == NULL) {
    perror("allocate args");
    exit(1);
  }

  if ((tv1 = (struct timeval*)malloc(sizeof(struct timeval))) == NULL) {
    perror("allocate tv1");
    exit(1);
  }

  if ((tv2 = (struct timeval*)malloc(sizeof(struct timeval))) == NULL) {
    perror("allocate tv2");
    exit(1);
  }
	
  if ((tz = (struct timezone*)malloc(sizeof(struct timezone))) == NULL) {
    perror("allocate tz");
    exit(1);
  }

  // Calculate the arguments (calargs.c)

  error = processArgs(argc, argv, args, rank);
      
  // If there is an error with the arguments print error and exit
  if (!rank) {
    if (error) {
      argErr();
    } else {
      // If argument is for verbose, print the operation 
      // and algorithm
      if (!(args->print)) {
        switch(args->type) {
        case(RING): {
          switch(args->op) {
          case(GATHER): {
            printf("Ring All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("Ring Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("Ring Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        case(FAN): {
          switch(args->op) {
          case(GATHER): {
            printf("Fan-in Fan-out All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("Fan-in Fan-out Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("Fan-in Fan-out Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        case(BIDIR): {
          switch(args->op) {
          case(GATHER): {
            printf("Bi-Directional All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("Bi-Directional Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("Bi-Directional Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        case(REC): {
          switch(args->op) {
          case(GATHER): {
            printf("Recursive-Halving All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("Recursive-Halving Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("Recursive-Halving Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        case(BUCK): {
          switch(args->op) {
          case(GATHER): {
            printf("Bucket All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("Bucket Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("Bucket Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        case(MPI): {
          switch(args->op) {
          case(GATHER): {
            printf("MPI All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("MPI Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("MPI Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        case(PIPE): {
          switch(args->op) {
          case(GATHER): {
            printf("Pipeline All-Gather\n");
            break;
          }
          case(REDUCE): {
            printf("Pipeline Reduce\n");
            break;
          }
          case(REDSCAT): {
            printf("Pipeline Reduce-Scatter\n");
            break;
          }
          }
          break;
        }
        } 
	// Extra information to print in verbose mode
        printf("Number of Elements: %d\nNumber of Iterations: %d\nType: %d\nOperation: %d\nNumber of Steps: %d\nStep Size: %d\nPrint?: %d\nPrint Result?: %d\nDo Computations?: %d\nElement in 10?: %d\nBlocking Send?: %d\nVerify Routine?: %d\n", args->numelem, args->numiter, args->type, args->op, args->numsteps, args->stepsize, args->print, args->output, args->compute, args->notpow, args->block,args->check);
      }
    }
  }
  
  // If argument requires verification run verify function to 
  // compare the result of my algorithm with MPI (verify.c).
  if (!error) {
    if (args->check) {
      error = verify(args, rank, size);
      if ((!args->print) && (!error)) {
        if (!rank) printf("Subroutine verified correct!\n");
      }
    }
  }  

  if (error) return error;
  
  // Set intial data for the nodes (either 10 or power of 2)
  done = FALSE;
  count = 0;
  if (args->notpow) {
    test = 10.0;
  } else {
    test = pow(2,rank);
  }

  // This loop is for increasing number of steps
  while(!done) {

    // allocate memory for send and receive buffer (different for
    // each operation
    if (args->op == 0) {
      if ((sendbuf = (double*)malloc(args->numelem*sizeof(double))) == NULL) {
	perror("allocate sendbuf");
	exit(1);
      }
      
      if ((recvbuf = (double*)malloc(args->numelem*size*sizeof(double))) 
	  == NULL) {
	perror("allocate recvbuf");
	exit(1);
      }
      
    } else {
      if ((sendbuf = (double*)malloc(args->numelem*sizeof(double))) == NULL) {
	perror("allocate sendbuf");
	exit(1);
      }
      if ((recvbuf = (double*)malloc(args->numelem*sizeof(double))) == NULL) {
	perror("allocate recvbuf");
	exit(1);
      }
    }   
    
    // Loop for number of iterations
    for (i=0;i<args->numiter;i++) {
      // Initialise send and receive buffer values
      if (args->op == 0) {
        for (j=0;j<args->numelem;j++) sendbuf[j] = test;
        for (j=0;j<args->numelem*size;j++) recvbuf[j] = -1.0;
      } else {
        for (j=0;j<args->numelem;j++) sendbuf[j] = test;
        for (j=0;j<args->numelem;j++) recvbuf[j] = -1.0;
      }
      
      if (args->op == 2) {
        numrecvelem = args->numelem / size;
        if ((recvcounts = (int*)malloc(size*sizeof(int))) == NULL) {
	  perror("allocate recvcounts");
	  exit(1);
	}
        for(j=0;j<size;j++) {
          recvcounts[j] = numrecvelem;
        } 
      }   

      // Start my timing and run the algorithms 

      MPI_Barrier(MPI_COMM_WORLD);
      if (!rank) gettimeofday(tv1, tz);
      
      switch(args->op) {
      case(0): {

        // do non-blocking
        switch(args->type) {
        case(0): {
          ring_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(1): {
          fan_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(2): {
          bidir_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(3): {
          rec_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(4): {
          buck_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(5): {
          mpi_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(6): {
          pipe_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(7): {
          bing_gather(sendbuf, recvbuf, args, rank, size); 
          break;
        }
	case(8): {
	  naive_gather(sendbuf, recvbuf, args, rank, size);
	  break;
	}
	case(9): {
	  fourway_gather(sendbuf, recvbuf, args, rank, size);
	  break;
	}
	}
	break;
      }
      
      case(1): {
        
        // do non-blocking
        switch(args->type) {
        case(0): {
          ring_reduce(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(1): {
          fan_reduce(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(2): {
          bidir_reduce(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(3): {
          rec_reduce(sendbuf, recvbuf, args, rank, size); 
          break;
        }
        case(4): {
	  buck_reduce(sendbuf, recvbuf, args, rank, size); 
	  break;
        }
        case(5): {
	  mpi_reduce(sendbuf, recvbuf, args, rank, size); 
	  break;
        }
        case(6): {
          pipe_reduce(sendbuf, recvbuf, args, rank, size); 
          break;
        }
	case(8): {
	  naive_reduce(sendbuf, recvbuf, args, rank, size); 
	  break;
        }
	case(9): {
	  fourway_reduce(sendbuf, recvbuf, args, rank, size);
	  break;
	}
        } 
	break;
      }
      
      case(2): {

        // do non-blocking
        switch(args->type) {
        case(0): {
          ring_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
        case(1): {
          fan_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
        case(2): {
          bidir_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
        case(3): {
          rec_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
        case(4): {
          buck_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
	  break;
        }
        case(5): {
          mpi_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
        case(6): {
          pipe_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
        case(7): {
          bing_redscat(sendbuf, recvbuf, args, rank, size, recvcounts); 
          break;
        }
	case(8): {
	  naive_redscat(sendbuf, recvbuf, args, rank, size, recvcounts);
	  break;
	}
	case(9): {
	  fourway_redscat(sendbuf, recvbuf, args, rank, size, recvcounts);
	  break;
	}
	}
	break;
      }
      }

      MPI_Barrier(MPI_COMM_WORLD);

      // Get the other timing and calculate time difference
      if (!rank) {
	gettimeofday(tv2, tz); 
	timediff = gettimediff(tv1, tv2);
	
	if (avgtime == -1) {
	  avgtime = timediff;
	} else {
	  avgtime = (avgtime + timediff) / 2;
	}
	if (mintime > timediff) mintime = timediff;
	if (maxtime < timediff) maxtime = timediff;
      }
    }  
    free(sendbuf);
    free(recvbuf);
    count++;

    // print results
    if (!rank) {
      if (!(args->print)) {
	printf("rank = %d, numelem = %d, avg = %ldus min = %ldus max = %ldus\n"
	       , rank, args->numelem, avgtime, mintime, maxtime);            
      } else {
	switch(args->printtype) {
	case(0) : {
	  if (args->op == 2) {
	    printf("%d %ld\n",(args->numelem)/size, mintime);
	  } else {
	    printf("%d %ld\n",args->numelem, mintime);
	  }
	  break;
	}
	case(1) : {
	  if (args->op == 2) {
	    printf("%d %ld\n",(args->numelem)/size, avgtime);
	  } else {
	    printf("%d %ld\n",args->numelem, avgtime);
	  }
	  break;
	} 
	case(2) : {
	  if (args->op == 2) {
	    printf("%d %ld\n",(args->numelem)/size, maxtime);
	  } else {
	    printf("%d %ld\n",args->numelem, maxtime);
	  }
	  break;
	} 
	}
      }
    }
    
    // Increase the element size and repeat program until 
    // timesteps is met
    if (count >= args->numsteps) {
      done = TRUE; 
    } else {
      args->numelem += args->stepsize;
      avgtime = -1;
      mintime = LONG_MAX;
      maxtime = -1;
    }
  }
  
  MPI_Finalize();
  return 0;
}
