#include "combine.h"
#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"
#include <time.h>

/* 
 * Verification Program
 *
 * Description: This program is used to verify the implementation by 
 * comparing the results with that of MPI. Input is a bunch of random numbers
 * and output returns whether the implementation concurs with MPI or not.
 *
 */

void vError(int rank) {
  printf("The subroutine has produced an error in process %d!\n", rank);
}

int verify(struct args *args, int rank, int size) {
  int error;
  double *rbuf1; // receive buffer for implementation
  double *rbuf2; // receive buffer for MPI
  double *sbuf; // sendbuffer (random values)
  int i; 
  int numrecvelem; // number of elements received
  int * recvcounts; // elements to be sent to each node

  error = FALSE; 
  srand((int)time); // create psuedo-random number based on time
  
  sbuf = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    sbuf[i] = rand();
    srand(sbuf[i]); // initalise send buffer
  }

  // checks for All-Gather algorithm to run based on commandline arguments

  switch(args->op) {
  case(GATHER) : {
    rbuf1 = (double*)malloc(args->numelem*size*sizeof(double));
    rbuf2 = (double*)malloc(args->numelem*size*sizeof(double));
    MPI_Allgather(sbuf, args->numelem, MPI_DOUBLE, rbuf1, 
                  args->numelem, MPI_DOUBLE, MPI_COMM_WORLD);
    switch(args->type) {
    case(RING) : {
      if (args->block) {
        b_ring_gather(sbuf, rbuf2, args, rank, size);
      } else {
        ring_gather(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(FAN) : {
      if (args->block) {
        b_fan_gather(sbuf, rbuf2, args, rank, size);
      } else {
        fan_gather(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(BIDIR) : {
      if (args->block) {
        b_bidir_gather(sbuf, rbuf2, args, rank, size);
      } else {
        bidir_gather(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(REC) : {
      if (args->block) {
        b_rec_gather(sbuf, rbuf2, args, rank, size);
      } else {
        rec_gather(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(BUCK) : {
      if (args->block) {
        b_buck_gather(sbuf, rbuf2, args, rank, size);
      } else {
        buck_gather(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(MPI) : {
      mpi_gather(sbuf, rbuf2, args, rank, size);
      break;
    }
    case(PIPE) : {
      if (args->block) {
        b_pipe_gather(sbuf, rbuf2, args, rank, size);
      } else {
        pipe_gather(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(BING) : {
      mpi_gather(sbuf, rbuf2, args, rank, size);
      break;
    }
    case(NAIVE) : {
      naive_gather(sbuf, rbuf2, args, rank, size);
      break;
    }
    }

    MPI_Barrier(MPI_COMM_WORLD);
    
    // compares MPI results with implementation.
    // Give an error if results are not the same.

    for (i=0;i<args->numelem*size;i++) {
      if (rbuf1[i] != rbuf2[i]) {
        if (!error) {
          printf("%f != %f\n",rbuf1[i],rbuf2[i]);
          vError(rank);
          error = TRUE;
        }
      }
    }
    
    break;
  }

    // Check for All-Reduce algorithm to run based on commandline arguments
    
  case(REDUCE) : {
    rbuf1 = (double*)malloc(args->numelem*sizeof(double));
    rbuf2 = (double*)malloc(args->numelem*sizeof(double));
    MPI_Allreduce(sbuf, rbuf1, args->numelem, MPI_DOUBLE, MPI_SUM, 
                MPI_COMM_WORLD);
    switch(args->type) {
    case(RING) : {
      if (args->block) {
        b_ring_reduce(sbuf, rbuf2, args, rank, size);
      } else {
        ring_reduce(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(FAN) : {
      if (args->block) {
        b_fan_reduce(sbuf, rbuf2, args, rank, size);
      } else {
        fan_reduce(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(BIDIR) : {
      if (args->block) {
        b_bidir_reduce(sbuf, rbuf2, args, rank, size);
      } else {
        bidir_reduce(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(REC) : {
      if (args->block) {
        b_rec_reduce(sbuf, rbuf2, args, rank, size);
      } else {
        rec_reduce(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(BUCK) : {
      if (args->block) {
        b_buck_reduce(sbuf, rbuf2, args, rank, size);
      } else {
        buck_reduce(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(MPI) : {
        mpi_reduce(sbuf, rbuf2, args, rank, size);
      
      break;
    }
    case(PIPE) : {
      if (args->block) {
        b_pipe_reduce(sbuf, rbuf2, args, rank, size);
      } else {
        pipe_reduce(sbuf, rbuf2, args, rank, size);
      }
      break;
    }
    case(NAIVE) : {
      naive_reduce(sbuf, rbuf2, args, rank, size);
      
      break;
    }
    } 

    MPI_Barrier(MPI_COMM_WORLD);
    
    // compares MPI results with implementation.
    // Give an error if results are not the same.

    for (i=0;i<args->numelem;i++) {
      if (rbuf1[i] != rbuf2[i]) {
        if (!error) {
          printf("%f != %f\n",rbuf1[i],rbuf2[i]);
          vError(rank);
          error = TRUE;
        }
      }
    }

    break;
  }

    // Check for Reduce-Scatter algorithm to run based on commandline argument 

  case(REDSCAT) : {
    rbuf1 = (double*)malloc(args->numelem*sizeof(double));
    rbuf2 = (double*)malloc(args->numelem*sizeof(double));
    numrecvelem = args->numelem / size;
    recvcounts = (int*)malloc(size*sizeof(int));

    for(i=0;i<size;i++) {
      recvcounts[i] = numrecvelem;
    }    

    MPI_Reduce_scatter(sbuf, rbuf1, recvcounts, MPI_DOUBLE, MPI_SUM, 
                       MPI_COMM_WORLD);
    switch(args->type) {
    case(RING) : {
      if (args->block) {
        b_ring_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      } else {
        ring_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      }
      break;
    }
    case(FAN) : {
      if (args->block) {
        b_fan_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      } else {
        fan_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      }
      break;
    }
    case(BIDIR) : {
      if (args->block) {
        b_bidir_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      } else {
        bidir_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      }
      break;
    }
    case(REC) : {
      if (args->block) {
        b_rec_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      } else {
        rec_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      }
      break;
    }
    case(BUCK) : {
      if (args->block) {
        b_buck_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      } else {
        buck_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      }
      break;
    }
    case(MPI) : {
        mpi_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      break;
    }
    case(PIPE) : {
      if (args->block) {
        b_pipe_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      } else {
        pipe_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      }
      break;
    }
    case(BING) : {
      bing_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      
      break;
    } 
    case(NAIVE) : {
      naive_redscat(sbuf, rbuf2, args, rank, size, recvcounts);
      
      break;
    }   
    }
    MPI_Barrier(MPI_COMM_WORLD);

    // compares MPI results with implementation.
    // Give an error if results are not the same.
    
    for (i=0;i<recvcounts[rank];i++) {
      if (rbuf1[i] != rbuf2[i]) {
        if (!error) {
          printf("%f != %f\n",rbuf1[i],rbuf2[i]);
          vError(rank);
          error = TRUE;
        }
      }
    }
  break;
  }
  }


  return error;
}
    
    
    
    





